# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 20th 2020

Modified By: hsky77
Last Updated: September 3rd 2020 11:18:07 am
'''


from . import RequestController, WebSocketController, StreamingDownloadController, StreamingUploadController
from ...constants import GB


class TestController(RequestController):
    def test_method(self, value: int):
        return value

    async def get(self):
        """
        ---
        tags:
        - test api
        summary: run for test base controller get
        description: base api test
        produces:
        - text/html
        responses:
            200:
              description: return test text
        """
        response = await self.component_services.invoke_async('test_www', 'get', params={'text': 'hello this is test'})
        self.write(response.text + '\n')

        response = await self.component_services.invoke_async('test_api', 'post', data={'value': 'hello this is test'})
        self.log_info(response.text)
        self.write(response.text + '\n')

        response = await self.component_services.invoke_async('test_api', 'put', data={'value': 'hello this is test'})
        self.log_info(response.text)
        self.write(response.text + '\n')

        response = await self.component_services.invoke_async('test_api', 'delete', params={'id': 'hello this is test'})
        self.log_info(response.text)
        self.write(response.text + '\n')

        with self.component_executor.get_executor() as executor:
            result = executor.run_method(self.test_method, 1)

        with self.component_executor.get_executor() as executor:
            result = await executor.run_method_async(self.test_method, 1)

    async def post(self):
        """
        ---
        tags:
        - test api
        summary: run for test base controller post
        description: base api test
        produces:
        - text/html
        responses:
            200:
              description: return test text
        """

        self.write('I got this post value: {}'.format(
            self.get_body_argument('value')))

    async def put(self):
        """
        ---
        tags:
        - test api
        summary: run for test base controller put
        description: base api test
        produces:
        - text/html
        responses:
            200:
              description: return test text
        """

        self.write('I got this put value: {}'.format(
            self.get_body_argument('value')))

    async def delete(self):
        """
        ---
        tags:
        - test api
        summary: run for test base controller delete
        description: base api test
        produces:
        - text/html
        responses:
            200:
              description: return test text
        """

        self.write('I got this id to delete: {}'.format(
            self.get_query_argument('id')))


class TestStreamDownloadController(StreamingDownloadController):
    async def _prepare_binary(self):
        self.set_header('Content-Disposition',
                        'attachment; filename=' + 'test.data')
        b = b''.ljust(1*GB, b'0')
        return b


class TestStreamUploadController(StreamingUploadController):
    async def prepare(self):
        self.buffer = []

    def _on_chunk_received(self, headers, chunk, bytes_size_received):
        self.buffer.append(chunk)

    def _on_data_received(self, headers, bytes_size_received):
        data = b''.join(self.buffer)
        if 'charset' in self.content_mime_options:
            data = data.decode(self.content_mime_options['charset'])
        self.write(data)


class TestWebsocketController(WebSocketController):
    live_web_sockets = set()

    def on_message(self, message):
        self.send_message(message)

    def open(self):
        self.set_nodelay(True)
        self.live_web_sockets.add(self)

    def on_close(self):
        self.live_web_sockets.remove(self)

    def send_message(self, message):
        removable = set()
        for ws in self.live_web_sockets:
            if not ws.ws_connection or not ws.ws_connection.stream.socket:
                removable.add(ws)
            else:
                ws.write_message(message)
        for ws in removable:
            self.live_web_sockets.remove(ws)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 20th 2020

Modified By: hsky77
Last Updated: September 2nd 2020 13:10:29 pm
'''

import os
import math
import asyncio
import cgi
from typing import Any, Dict, Union

from tornado.ioloop import IOLoop
from tornado.web import RequestHandler, stream_request_body
from tornado.websocket import WebSocketHandler

from .mixin import TornadoMixin

from ....util import join_path, FunctionQueueWorker
from ...constants import GB, KB, LocalCode_Missing_Required_Parameter, LocalCode_Upload_Success


class RequestController(TornadoMixin, RequestHandler):
    """base http request hanlder class of hyssop"""

    def set_default_headers(self):
        self.set_header('App-Name', self.application_name)


class WebSocketController(TornadoMixin, WebSocketHandler):
    """base web socket hanlder class of hyssop"""

    def __init__(self, application, request, **kwds):
        super().__init__(application, request, **kwds)
        self.io_loop = IOLoop.current(False)

    def set_default_headers(self):
        self.set_header('App-Name', self.application_name)

    def write_message(self, message: Union[bytes, str, Dict[str, Any]], binary: bool = False) -> 'Future[None]':
        if IOLoop.current(False) is self.io_loop:
            try:
                return super().write_message(message, binary)
            except IOError:
                self.io_loop.add_callback(self.on_close)
        else:
            self.io_loop.add_callback(self.write, message, binary)


class StreamingDownloadController(RequestController):
    SUPPORTED_METHODS = ("GET", "OPTIONS")

    def initialize(self, chunk_size: int = 4*KB, **kwds):
        self.chunk_size = chunk_size

    async def _prepare_binary(self):
        raise NotImplementedError()

    async def get(self):
        bdata = await self._prepare_binary()
        self.set_header('Content-Type', 'application/octet-stream')

        for i in range(0, math.ceil(len(bdata) / self.chunk_size)):
            start = i * self.chunk_size
            data = bdata[start:min(start+self.chunk_size, len(bdata))]
            self.write(data)
            await asyncio.sleep(0)

        self.finish()


@stream_request_body
class StreamingUploadController(RequestController):
    SUPPORTED_METHODS = ("POST", "PUT", "OPTIONS")

    def initialize(self, upload_dir: str = '', max_stream_size: int = 1*GB,  **kwds):
        self.upload_dir = join_path(self.root_dir, upload_dir)
        self.request.connection.set_max_body_size(max_stream_size)

        self._data_worker = FunctionQueueWorker()
        self.content_mime_type, self.content_mime_options = cgi.parse_header(
            self.request.headers['Content-Type'])

        self.bytes_length = 0

    async def prepare(self):
        pass

    async def data_received(self, chunk):
        self.bytes_length = self.bytes_length + len(chunk)
        self._data_worker.run_method(
            self._on_chunk_received, self.request.headers, chunk, self.bytes_length)

    async def post(self):
        while not self._data_worker.pending_count == 0:
            await asyncio.sleep(0)
        self._data_worker.dispose()

        with self.component_executor.get_executor() as executor:
            await executor.run_method_async(self._on_data_received, self.request.headers, self.bytes_length)

    async def put(self):
        while not self._data_worker.pending_count == 0:
            await asyncio.sleep(0)
        self._data_worker.dispose()

        with self.component_executor.get_executor() as executor:
            await executor.run_method_async(self._on_data_received, self.request.headers, self.bytes_length)

    def _on_chunk_received(self, headers, chunk, bytes_size_received):
        raise NotImplementedError()

    def _on_data_received(self, headers, bytes_size_received):
        pass

    def _handle_request_exception(self, e: BaseException) -> None:
        self._data_worker.dispose()
        super()._handle_request_exception(e)


class StreamingFileUploadController(StreamingUploadController):
    async def prepare(self):
        await super().prepare()
        self.filename = self.request.headers.get("filename")
        if self.filename is None:
            raise KeyError(self.component_localization.get_message(
                LocalCode_Missing_Required_Parameter, 'filename'))

        path = join_path(self.upload_dir, self.filename)

        if not os.path.isdir(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        self.__file = open(path, 'wb')

    def _on_chunk_received(self, headers, chunk, bytes_size_received):
        if self.__file is not None:
            self.__file.write(chunk)

    def _on_data_received(self, headers, bytes_size_received):
        if self.__file is not None:
            self.__file.close()
            self.__file = None

            self.log_info(self.component_localization.get_message(
                LocalCode_Upload_Success, self.filename))
            self.write(self.component_localization.get_message(
                LocalCode_Upload_Success, self.filename))



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: September 5th 2020 20:56:18 pm
'''

import os
import sys
from .util import get_class, join_path
from . import Module_Path, Version


def start_server(server_dir: str, cmdline_args) -> None:
    from .util import get_class

    try:
        route = server_dir + '.server'
        cls = server_dir + '_server'
        server_cls = get_class(route, cls)
    except:
        from .web.server import HyssopServer
        server_cls = HyssopServer

    server = server_cls()
    server.init(server_dir, cmdline_args.http_server)
    server.start()


def pack_server(server_dir: str, cmdline_args) -> None:
    from .web.pack import HyssopPack
    HyssopPack().pack(server_dir, cmdline_args.o,
                      prepare_wheels=cmdline_args.add_wheels,
                      compile_py=not cmdline_args.decompile_pyc)


def test_server():
    import unittest
    from .unit_test import get_test_suite
    runner = unittest.TextTestRunner()
    runner.run(get_test_suite())


def create_server_folder(server_dir: str = 'hello_world'):
    from .web import (Component_Module_Folder,
                      Controller_Module_Folder,
                      Unittest_Module_Folder,
                      Hyssop_Web_Config_File,
                      Hyssop_Web_Pack_File,
                      Component_Module_Folder,
                      Controller_Module_Folder,
                      Hyssop_Web_Requirement_File)
    from .web.server import HyssopServer

    if not os.path.isdir(server_dir):
        os.makedirs(server_dir)

    # create hello component
    comp_dir = join_path(server_dir, Component_Module_Folder)
    if not os.path.isdir(comp_dir):
        os.makedirs(comp_dir)

    with open(join_path(comp_dir, '__init__.py'), 'w') as f:
        f.write('''\
from hyssop.web.component import ComponentTypes

from hyssop.web.config_validator import ConfigContainerMeta, ConfigElementMeta, WebConfigComponentValidator

# add hello validator to component config validator
WebConfigComponentValidator.set_cls_parameters(
    ConfigContainerMeta('hello', False,
        ConfigElementMeta('p1', str, True) # validate HelloComponent's 'p1' argument is required and string type
    )
)

class HelloComponentTypes(ComponentTypes):
    Hello = ('hello', 'hello', 'HelloComponent')
''')

    with open(join_path(comp_dir, 'hello.py'), 'w') as f:
        f.write('''\
from hyssop.web.component import Component
from . import HelloComponentTypes

class HelloComponent(Component):
    def init(self, component_manager, p1, *arugs, **kwargs) -> None:
        print('init Hello component load from', __package__, 'and the parameters p1:', p1)

    def hello(self):
        return 'Hello World, This is hyssop generate hello component'
''')

    # create hello controller
    controller_dir = join_path(server_dir, Controller_Module_Folder)
    if not os.path.isdir(controller_dir):
        os.makedirs(controller_dir)

    with open(join_path(controller_dir, '__init__.py'), 'w') as f:
        f.write('''\
from hyssop.web.controller import ControllerType

class HelloControllerTypes(ControllerType):
    HelloController = ('hello_world', 'hello', 'HelloController')
    THelloController = ('tornado_hello_world', 'hello', 'HelloTornadoHandler')
''')

    with open(join_path(controller_dir, 'hello.py'), 'w') as f:
        f.write('''\
from hyssop.web.controller.tornado import RequestController

from component import HelloComponentTypes

class HelloController(RequestController):
    async def get(self):
        """
        ---
        tags:
        - hello
        summary: hello world get
        description: simple test controller
        produces:
        - text/html
        responses:
            200:
              description: return hello message
        """

        hello_comp = self.component_manager.get_component(HelloComponentTypes.Hello)
        self.write(hello_comp.hello())

from tornado.web import RequestHandler

class HelloTornadoHandler(RequestHandler):
    async def get(self):
        """
        ---
        tags:
        - tornado hello
        summary: hello world get
        description: simple test controller
        produces:
        - text/html
        responses:
            200:
              description: return hello message
        """        

        hello_comp = self.application.component_manager.get_component(HelloComponentTypes.Hello)
        self.write(hello_comp.hello())
''')

    # unit test
    ut_dir = join_path(server_dir, Unittest_Module_Folder)
    if not os.path.isdir(ut_dir):
        os.makedirs(ut_dir)

    with open(join_path(ut_dir, '__init__.py'), 'w') as f:
        f.write('''\
from hyssop.unit_test import UnitTestTypes

class UTTypes(UnitTestTypes):
    UT1 = ('ut1', 'ut1', 'UT1TestCase')
''')

    with open(join_path(ut_dir, 'ut1.py'), 'w') as f:
        f.write('''\
from hyssop.unit_test import UnitTestCase

class UT1TestCase(UnitTestCase):
    def test(self):
        # implement unit test here...
        
        print('ut1 test case runned')
''')

    # config
    with open(join_path(server_dir, Hyssop_Web_Config_File), 'w') as f:
        f.write('''\
name: hyssop Server
port: 8888
debug: False
doc:
  description: hello api
component:
  hello: 
    p1: 'This is p1'
controller:
  /hello:
    enum: hello_world
  /hello_tornado:
    enum: tornado_hello_world
''')

    # pack
    with open(join_path(server_dir, Hyssop_Web_Pack_File), 'w') as f:
        f.write('''
# This is packing list indicated what are the files should be pack
# If this file does not exist under the project folder, all of the files under the folder will be packed

include:
# List absolute or relative path of additional file or directory to be packed
# - example.txt
# - example_dir

exclude:
# List absolute or relative path of file, directory, or file extension to be ignored.
- '.log'
''')

    # requirement
    from . import __name__, Version
    with open(join_path(server_dir, Hyssop_Web_Requirement_File), 'w') as f:
        f.write('{}>={}'.format(__name__, Version))

    print('project created at', os.path.abspath(server_dir))


Command_Start_Server = 'start'
Command_Pack_Server = 'pack'
Command_Test_Server = 'test'
Command_Create_Project = 'create'
Command_Show_Version = 'version'


def get_command_parser():
    import argparse
    parser = argparse.ArgumentParser(prog='hyssop')
    sub_parsers = parser.add_subparsers(title='command')

    start_parser = sub_parsers.add_parser(
        Command_Start_Server, help='start server with specfied server project directory path')
    start_parser.add_argument('server_directory',
                              help='server project directory path')
    start_parser.add_argument('-s', '--http_server', action='store_true',
                              help='start application on tornado http server')
    start_parser.set_defaults(command=Command_Start_Server)

    pack_parser = sub_parsers.add_parser(
        Command_Pack_Server, help='pack server with specfied server project directory path')
    pack_parser.add_argument('server_directory',
                             help='server project directory path')
    pack_parser.add_argument(
        '-o', help='specify output compressed file path', default=None)
    pack_parser.add_argument('-w', '--add_wheels', action='store_true',
                             help='add dependency wheel files')
    pack_parser.add_argument('-d', '--decompile_pyc', action='store_true',
                             help='disable compile .py to .pyc')
    pack_parser.set_defaults(command=Command_Pack_Server)

    test_parser = sub_parsers.add_parser(
        Command_Test_Server, help='test hyssop library or specfied server project directory path')
    test_parser.add_argument('server_directory', nargs='?',
                             help='server project directory path')
    test_parser.set_defaults(command=Command_Test_Server)

    make_serv_parser = sub_parsers.add_parser(
        Command_Create_Project, help='create a server template with specfied server project directory path')
    make_serv_parser.add_argument('server_directory',
                                  help='server project directory path')
    make_serv_parser.set_defaults(command=Command_Create_Project)

    version_serv_parser = sub_parsers.add_parser(
        Command_Show_Version, help='print version number to console')
    version_serv_parser.set_defaults(command=Command_Show_Version)

    return parser


def _init_server_package(server_dir: str) -> None:
    if server_dir is not None:
        if (not join_path(Module_Path) in server_dir and
                not os.path.dirname(os.path.abspath(server_dir)) in sys.path):
            sys.path.insert(0, os.path.dirname(os.path.abspath(server_dir)))

        if not join_path(Module_Path) in server_dir and not server_dir in sys.path:
            sys.path.insert(0, os.path.abspath(server_dir))

        from .util import get_class
        from .web import Component_Module_Folder, Controller_Module_Folder, Unittest_Module_Folder
        try:
            get_class(Component_Module_Folder)
        except:
            pass

        try:
            get_class(Controller_Module_Folder)
        except:
            pass

        try:
            get_class(Unittest_Module_Folder)
        except:
            pass


if __name__ == '__main__':
    parser = get_command_parser()
    args = parser.parse_args()

    if hasattr(args, 'command'):
        if args.command == Command_Start_Server:
            server_dir = args.server_directory
            _init_server_package(server_dir)
            start_server(server_dir, args)
        elif args.command == Command_Pack_Server:
            from .util import join_to_abs_path
            pack_server(join_to_abs_path(args.server_directory), args)
        elif args.command == Command_Test_Server:
            server_dir = args.server_directory
            _init_server_package(server_dir)
            test_server()
        elif args.command == Command_Create_Project:
            create_server_folder(args.server_directory)
        elif args.command == Command_Show_Version:
            print('hyssop {}'.format(Version))
        else:
            parser.print_help()
    else:
        parser.print_help()



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 20th 2020

Modified By: hsky77
Last Updated: August 27th 2020 16:19:09 pm
'''


import time
import asyncio

from ..web import Component_Module_Folder, Controller_Module_Folder
from .base import UnitTestCase, UnitTestServer
from ..util import join_path


class WebTestCase(UnitTestCase):
    def test(self):
        self.test_components()
        self.test_server()

    def test_server(self):
        from .. import Module_Path
        from ..web.component import DefaultComponentTypes, create_server_component_manager
        from ..web.component.default import ServicesComponent

        # start a test server
        dir_path = join_path(Module_Path, 'web')
        server = UnitTestServer()

        component_setting = {
            'services':
            {
                'routes': {
                    'https://www.google.com': {
                        'test_www': '/'
                    },
                    'http://localhost:58564': {
                        'test_api': '/test_api'
                    }
                }
            }
        }
        server.set_config(dir_path,
                          component_config=component_setting,
                          controller_config={
                              '/test_api': {
                                  'enum': 'test_api'},
                              '/test_download': {
                                  'enum': 'test_download'
                              },
                              '/test_socket': {
                                  'enum': 'test_socket'
                              },
                              '/test_bytes_upload': {
                                  'enum': 'test_bytes_upload'
                              }
                          })

        server.start()

        component_manager = create_server_component_manager(
            component_setting, dir_path)

        service: ServicesComponent = component_manager.get_component(
            DefaultComponentTypes.Service)

        try:
            # test request
            response = None
            while not response:
                if server.server_exception:
                    raise server.server_exception
                try:
                    # pass
                    response = service.invoke(
                        'http://localhost:58564/test_api')
                except:
                    time.sleep(0.5)

            self.assertEqual(response.status_code, 200)

            # test bytes upload
            from ..web.application.tornado import StreamUploadClient
            client = StreamUploadClient()
            response = client.send_bytes(
                'http://localhost:58564/test_bytes_upload', b'i am bytes')
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.text, 'i am bytes')

            # # test web socket
            from ..web.application.tornado import WebSocketClient
            with WebSocketClient() as client:
                self.received_msg = None

                def received_ws_msg(msg):
                    self.received_msg = msg

                client.connect('ws://localhost:58564/test_socket',
                               on_message_callback=received_ws_msg)
                client.write_message('Hello There')

                while self.received_msg is None:  # wait for message come back
                    time.sleep(0)

                self.assertEqual(self.received_msg, 'Hello There')

        finally:
            server.stop()

    def test_components(self):
        import os
        from ..web.component import create_server_component_manager, DefaultComponentTypes
        from ..web.component.default import LocalizationComponent, LoggerComponent, CallbackComponent, ExecutorComponent, ServicesComponent
        from ..web.constants import LocalCode_Application_Closing
        from .. import Module_Path
        from ..util import join_path
        dir_path = join_path(Module_Path, 'web')

        component_setting = {
            'services':
            {
                'routes': {
                    'https://www.google.com': {
                        'test_www': '/'
                    }
                }
            }
        }

        component_manager = create_server_component_manager(
            component_setting, dir_path)

        self.test_func_runned = False

        try:
            # test default components:
            # localization
            local: LocalizationComponent = component_manager.get_component(
                DefaultComponentTypes.Localization)
            local.set_language('en')
            self.assertEqual(local.get_message(
                LocalCode_Application_Closing), 'stopping application')

            # log
            log: LoggerComponent = component_manager.get_component(
                DefaultComponentTypes.Logger)
            logger = log.get_logger('test_logger')
            logger.info('Test Message')
            for path in log.default_loggers + ['test_logger']:
                path = join_path(dir_path, 'files', 'logs', path + '.log')
                self.assertFalse(os.path.isfile(path))

            # callback
            from .util import UtilTestCase, TestCallbackType
            callbacks: CallbackComponent = component_manager.get_component(
                DefaultComponentTypes.Callback)
            UtilTestCase.test_callback(
                self, callbacks.get_callback_obj(TestCallbackType))

            # executor
            executor: ExecutorComponent = component_manager.get_component(
                DefaultComponentTypes.Executor)
            executor.run_method(UtilTestCase.test_func, self, 1, kwindex=2)
            loop = asyncio.get_event_loop()
            loop.run_until_complete(executor.run_method_async(
                UtilTestCase.test_func, self, 1, kwindex=2))

            def test_func(result):
                self.test_func_runned = True

            executor.run_method_in_queue(UtilTestCase.test_func, self, 1,
                                         kwindex=2, on_finish=test_func)

            # services
            services: ServicesComponent = component_manager.get_component(
                DefaultComponentTypes.Service)
            response = services.invoke('test_www')
            self.assertEqual(response.status_code, 200)

        finally:
            # queue will run at this moment if this test runs too fast
            loop.run_until_complete(
                component_manager.boardcast_async('dispose', component_manager))
            self.assertTrue(self.test_func_runned)



# Copyright (C) 2019-Present the hyssop authors and contributors
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php:

'''
server loads all default components in hyssop.web.component.DefaultComponentTypes when starting

    LocalizationComponent:

        - managing localized message via config setting such as:

        component:
            localization:
                dir: 'local'    # optional: Load the .csv files in the directory under project directory.
                lang: 'en'      # optional: Setup language, default: en.

    LoggerComponent:

        - managing logger via config setting such as:

        component:
            logger:
                log_to_resources: False     # optional: Enable log to resources
                log_to_console: False       # optional: Enable log to console
                dir: 'logs'                 # optional. Log to files in the folder under porject directory if specified

    CallbackComponent:

        - managing callback function by enum types, no configuration settings needed

    ExecutorComponent:

        - managing executor to execute the given functions and arguments

        component:
            executor:
                worker_count: 1      # The maximum of workers is 2

    ServicesComponent:

        - managing url route and service apis and provide inovke methons:

        component:
            services:
                async_connection_limit:             <int>   # the connections limitation in async mode
                async_connection_limit_pre_host:    <int>   # the connections limitation of each host in async mode
                routes:
                    url:                            <str>
                        name:   api_route           <str>
                    url:                            <str>
                        etc...

Last Updated: September 4th 2020 14:27:50 pm
'''

import logging
import requests

from datetime import datetime
from enum import Enum
from typing import Callable, Any, List, Dict, ByteString

from aiohttp import ClientResponse, TCPConnector, ClientSession

from . import Component, ComponentManager, ComponentTypes
from ...util import ExecutorFactory, Executor, Callbacks, configure_colored_logging, join_path, BaseSyncLogger
from .mixin import FileLoggerMixin


class LocalizationComponent(Component):
    """default component for managing localized message by config setting"""

    def init(self, component_manager: ComponentManager, lang: str = 'en', **kwargs) -> None:
        import os
        from hyssop.util import BaseLocal
        self.local = BaseLocal

        self.dir = kwargs.get('dir', None)
        if self.dir is not None:
            self.dir = join_path(kwargs.get('root_dir', ''), self.dir)
            self.local.import_csvs_from_directory(self.dir)

        if lang is not None:
            self.local.set_language(lang)

    def info(self) -> str:
        return {**super().info(), **{'info': self.local.get_info()}}

    @property
    def current_language(self) -> str:
        return self.local.current_language

    def set_language(self, lang: str) -> None:
        self.local.set_language(lang)

    def get_message(self, code: str, *args) -> str:
        """convert to localized message with code and following parameters"""
        return self.local.get_message(code, *args)


class LoggerComponent(FileLoggerMixin, Component):
    """default component for managing logger by server config"""

    default_loggers = []

    def __init__(self, component_type: ComponentTypes):
        super().__init__(component_type)
        configure_colored_logging()
        self.loggers = {}
        self.kwargs = None

    def init(self, component_manager: ComponentManager, **kwargs) -> None:
        self.kwargs = kwargs
        self.log_level = logging.INFO
        self.log_echo = self.kwargs.get('log_to_console', False)
        self.log_to_resources = self.kwargs.get('log_to_resources', False)

    def get_logger(self, name: str, *args, sub_dir: str = '', mode: str = 'a', encoding: str = 'utf-8', echo: bool = False) -> BaseSyncLogger:
        """create and return logger object, sub_dir appends the path to configured log path"""
        logger = self.loggers[name] if name in self.loggers else None

        if not logger:
            logger = logging.getLogger(name)

            if self.log_to_resources:
                self.update_file_handler(
                    logger, sub_dir, mode, encoding, **self.kwargs)
            else:
                self.remove_file_handler(logger, sub_dir, **self.kwargs)

            self.loggers[name] = logger

        if logger:
            logger.setLevel(self.log_level)
            logger.propagate = self.log_echo or echo

        return logger

    def update_default_logger(self, debug: bool = False) -> None:
        self.log_level = logging.DEBUG if debug else self.log_level

        for name in self.default_loggers:
            self.get_logger(name)

        self.loggers.clear()


class CallbackComponent(Component):
    """default component for managing callback function by enum types"""

    def init(self, component_manager: ComponentManager, *arugs, **kwargs) -> None:
        self._callback_manager = {}

    def get_callback_obj(self, enum_cls: Enum) -> Callbacks:
        if not enum_cls in self._callback_manager:
            self._callback_manager[enum_cls] = Callbacks(enum_cls)
        return self._callback_manager[enum_cls]

    def add_callback(self, callback_enum_type: Enum, callback: Callable) -> None:
        if isinstance(callback_enum_type, Enum):
            enum_cls = type(callback_enum_type)
            if not enum_cls in self._callback_manager:
                self._callback_manager[enum_cls] = Callbacks(enum_cls)

            self._callback_manager[enum_cls].add_callback(
                callback_enum_type, callback)

    def remove_callback(self, callback_enum_type: Enum, callback: Callable) -> None:
        if isinstance(callback_enum_type, Enum):
            enum_cls = type(callback_enum_type)
            if enum_cls in self._callback_manager:
                self._callback_manager[enum_cls].remove_callback(
                    callback_enum_type, callback)

    def execute_callback(self, callback_enum_type: Enum, *args, **kwargs) -> None:
        from tornado.ioloop import IOLoop
        if isinstance(callback_enum_type, Enum):
            enum_cls = type(callback_enum_type)
            if enum_cls in self._callback_manager:
                self._callback_manager[enum_cls].execute_callback(
                    callback_enum_type, *args, **kwargs)

    async def execute_callback_async(self, callback_enum_type: Enum, *args, **kwargs) -> None:
        if isinstance(callback_enum_type, Enum):
            enum_cls = type(callback_enum_type)
            if enum_cls in self._callback_manager:
                await self._callback_manager[enum_cls].execute_callback_async(
                    callback_enum_type, *args, **kwargs)


class ExecutorComponent(Component):
    def init(self, component_manager: ComponentManager, worker_count: int = 1, **kwargs) -> None:
        self.executor_factory = ExecutorFactory(worker_limit=worker_count)
        self.disposing = False

    def info(self) -> Dict:
        return {**super().info(), **{
            'info': {
                'workers': self.executor_factory.worker_count
            }
        }}

    def run_method_in_queue(self,
                            func: Callable,
                            *args,
                            on_finish: Callable[[Any], None] = None,
                            on_exception: Callable[[Exception], None] = None,
                            **kwargs) -> None:
        if not self.disposing:
            self.executor_factory.run_method_in_queue(
                func, *args, on_finish=on_finish, on_exception=on_exception, **kwargs)

    def run_method(self, func: Callable, *args, **kwargs) -> Any:
        if not self.disposing:
            return self.executor_factory.run_method(func, *args, **kwargs)

    async def run_method_async(self, func: Callable, *args, **kwargs) -> Any:
        if not self.disposing:
            return await self.executor_factory.run_method_async(func, *args, **kwargs)

    def get_executor(self) -> Executor:
        return self.executor_factory.get_executor()

    def dispose(self, component_manager: ComponentManager) -> None:
        self.disposing = True
        self.executor_factory.dispose()


class ServicesComponent(Component):
    """default component for managing url route and service apis"""

    STREAMING_CHUNK_SIZE = 8192

    @property
    def async_client(self) -> ClientSession:
        if not hasattr(self, 'aclient'):
            self.aclient = ClientSession(connector=TCPConnector(
                limit=self.async_connection_limit, limit_per_host=self.async_connection_limit_pre_host))
        return self.aclient

    def init(self, component_manager: ComponentManager, **kwargs) -> None:
        self.async_connection_limit = kwargs.get('async_connection_limit', 30)
        self.async_connection_limit_pre_host = kwargs.get(
            'async_connection_limit_pre_host', 10)

        self.routes = {}
        routes = kwargs.get('routes', {})
        for url, apis in routes.items():
            for name, api in apis.items():
                self.routes[name] = '{}{}'.format(url, api)

    def invoke(self,
               service_name_or_url: str,
               method: str = 'get',
               sub_route: str = '',
               streaming_callback: Callable = None,
               chunk_size: int = STREAMING_CHUNK_SIZE,
               **kwargs) -> requests.Response:
        """
        This function wraps requests.request(). That means this function accepts the same parameters as requests.request()

        Note: 
            use params= {} to send query parameters when method is 'get' or 'delete'
            use data= {} to send body parameters when method is the others
        """
        url = self.routes[service_name_or_url] if service_name_or_url in self.routes else service_name_or_url

        if not sub_route == '' and not sub_route == None:
            url = '{}/{}'.format(url, sub_route)

        if callable(streaming_callback):
            with requests.request(method, url, stream=True, **kwargs) as response:
                response.raise_for_status()
                for chunk in response.iter_content(chunk_size=chunk_size):
                    streaming_callback(chunk)
                return response
        else:
            return requests.request(method, url, **kwargs)

    async def invoke_async(self,
                           service_name_or_url: str,
                           method: str = 'get',
                           sub_route: str = '',
                           streaming_callback: Callable[[bytes], None] = None,
                           chunk_size: int = STREAMING_CHUNK_SIZE,
                           **kwargs) -> requests.Response:
        """
        This function wraps aiohttp.ClientSession.request(). That means this function accepts the same parameters as aiohttp.ClientSession.request().
        The returned response is requests.Response to allow the similar usage of the response instance as self.invoke()

        Note: 
            use params= {} to send query parameters when method is 'get' or 'delete'
            use data= {} to send body parameters when method is the others
        """

        url = self.routes[service_name_or_url] if service_name_or_url in self.routes else service_name_or_url

        if not sub_route == '' and not sub_route == None:
            url = '{}/{}'.format(url, sub_route)

        result = requests.Response()
        dt = datetime.now()
        if callable(streaming_callback):
            async with self.async_client.request(method, url, **kwargs) as response:
                async for chunk in response.content.iter_chunked(chunk_size):
                    streaming_callback(chunk)
        else:
            async with self.async_client.request(method, url, **kwargs) as response:
                result._content = await response.read()

        result.status_code = response.status
        result.url = str(response.real_url)
        result.headers = {k: v for k, v in response.headers.items()}
        result.cookies = response.cookies
        result.history = response.history
        result.elapsed = datetime.now() - dt
        result.encoding = response.get_encoding()
        result.request = response.request_info
        return result

    async def dispose(self, component_manager: ComponentManager):
        if hasattr(self, 'aclient') and self.aclient:
            await self.aclient.close()



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 20th 2020

This module provicde the classes integrate tornado application and clients.

    - TornadoApplication: add component-based architecture to tornado application

    - WebSocketClient communicates with server in web socket, for example:

        with WebSocketClient() as c:
            c.connect('ws://localhost:8888/test_socket', lambda msg: print('received message: {}'.format(msg)))
            text = input().strip()
            while text is not 'q':
                c.write_message(text)
                text = input().strip()

    - StreamUploadClient is the data uploader via streaming http request. for example:

        with StreamUploadClient() as c:

            # send file
            c.send_file('http://localhost:8888/test_file_upload', 'test.py', 'test.py')

            # send some bytes
            c.send_bytes('http://localhost:8888/test_bytes_upload', b'000000000  weftest.py')

            # send string
            c.send_string('http://localhost:8888/test_bytes_upload', '000000000  weftest.py')

            # send with generator
            def gen():
                for i in range(20):
                    yield str(i).encode('utf-8')

            c.send_by_generator('http://localhost:8888/test_bytes_upload', gen(), encoding='utf-8')

            # async
            loop = asyncio.new_event_loop()

            loop.run_until_complete(c.send_file_async(
                'http://localhost:8888/test_file_upload', 'test.py', 'test.py'))

            loop.run_until_complete(c.send_bytes_async(
                'http://localhost:8888/test_bytes_upload', b'wefwefwegge  weftest.py'))
            
            loop.run_until_complete(c.send_string_async(
                'http://localhost:8888/test_bytes_upload', 'wefwefwegge  weftest.py'))

            loop.run_until_complete(c.send_by_generator_async(
                'http://localhost:8888/test_bytes_upload', gen(), encoding='utf-8'))

Modified By: hsky77
Last Updated: September 4th 2020 14:23:47 pm
'''

import os
import time
import asyncio
import requests
import logging

from typing import Callable, ByteString, Generator, Union, Dict, Any

from tornado.ioloop import IOLoop
from tornado.web import Application

from ...util import FunctionLoopWorker, BaseLocal
from ..constants import LocalCode_Connect_Failed, KB
from .base import ComponentApplicationMixin


class TornadoApplication(ComponentApplicationMixin, Application):

    def init_application_mixin(self, settings: Dict[str, Any]) -> None:
        from ..component import add_module_default_logger

        # add tornado logger to the default loggers of LoggerComponent
        add_module_default_logger(
            ['tornado.access', 'tornado.application', 'tornado.general'])

        super().init_application_mixin(settings)

        self.logger = logging.getLogger('tornado.application')


class WebSocketClient(FunctionLoopWorker):
    def __init__(self, name: str = None, reconnect: int = 3):
        super().__init__(name, 0)
        self.conn = None
        self.reconnect = reconnect

    def dispose(self):
        self.disconnect()
        super().dispose()

    @property
    def is_connected(self):
        return self.conn is not None

    def connect(self, url: str, on_message_callback: Callable = None):
        if self.conn is None:
            self.url = url
            self.connect_count = 0
            self.on_message_callback = on_message_callback
            self.run_method(self._start)

            while not self.is_connected and self.connect_count < self.reconnect:  # wait for connected
                import time
                time.sleep(0)

            if not self.is_connected:
                raise RuntimeError(BaseLocal.get_message(
                    LocalCode_Connect_Failed, self.url))

    async def connect_async(self, url: str, on_message_callback: Callable = None):
        if self.conn is None:
            self.url = url
            self.connect_count = 0
            self.on_message_callback = on_message_callback
            self.run_method(self._start)

            while not self.is_connected and self.connect_count < self.reconnect:  # wait for connected
                import asyncio
                await asyncio.sleep(0)

            if not self.is_connected:
                raise RuntimeError(BaseLocal.get_message(
                    LocalCode_Connect_Failed, self.url))

    def disconnect(self):
        if self.conn is not None:
            if IOLoop.current(False) is self.io_loop:
                try:
                    self.conn.close()
                    self.conn = None
                except IOError:
                    pass
                finally:
                    self.stop()
            else:
                self.io_loop.add_callback(self.disconnect)

    def write_message(self, message: Union[str, bytes], binary: bool = False) -> None:
        if self.conn is not None:
            if IOLoop.current(False) is self.io_loop:
                try:
                    self.conn.write_message(message)
                except IOError:
                    self.io_loop.add_callback(self.disconnect)
            else:
                self.io_loop.add_callback(
                    self.conn.write_message, message, binary)

    def _start(self):
        import asyncio
        if self.connect_count < self.reconnect:
            try:
                loop = asyncio.new_event_loop()
                loop.run_until_complete(self.__run())
            finally:
                self.connect_count = self.connect_count + 1

    async def __run(self):
        from tornado.websocket import websocket_connect
        self.conn = await websocket_connect(self.url)
        self.io_loop = IOLoop.current(False)
        msg = await self.conn.read_message()
        while msg is not None:
            if callable(self.on_message_callback):
                self.on_message_callback(msg)
            msg = await self.conn.read_message()


class StreamUploadClient():
    def send_file(self, url: str, source: str, target: str, chunk_size: int = 4*KB,
                  on_chunk_send: Callable[[int, int], None] = None) -> requests.Response:
        return self._send_file(url, source, target, chunk_size, on_chunk_send)

    def _send_file(self, url: str, source: str, target: str, chunk_size: int = 4*KB,
                   on_chunk_send: Callable[[int, int], None] = None) -> requests.Response:
        statinfo = os.stat(source)
        headers = {"Content-Type": "application/octet-stream",
                   "filename": target}
        with open(source, 'rb') as f:
            response = requests.put(url, data=self.__send(
                f, chunk_size, on_chunk_send, statinfo.st_size), headers=headers)
            response.close()
            return response

    def send_bytes(self, url: str, data: ByteString, chunk_size: int = 4*KB,
                   on_chunk_send: Callable[[int, int], None] = None, encoding: str = None) -> requests.Response:
        return self._send_bytes(url, data, chunk_size, on_chunk_send, encoding)

    def send_string(self, url: str, data: str, chunk_size: int = 4*KB,
                    on_chunk_send: Callable[[int, int], None] = None, encoding: str = 'utf-8') -> requests.Response:
        return self.send_bytes(url, data.encode(encoding), chunk_size, on_chunk_send, encoding)

    def _send_bytes(self, url: str, data: ByteString, chunk_size: int = 4*KB,
                    on_chunk_send: Callable[[int, int], None] = None, encoding: str = None) -> requests.Response:
        from io import BytesIO
        content_type = "application/octet-stream"
        if encoding is not None:
            content_type = content_type + "; charset={}".format(encoding)

        headers = {"Content-Type": content_type}
        response = requests.post(url, data=self.__send(
            BytesIO(data), chunk_size, on_chunk_send, len(data)), headers=headers)
        response.close()
        return response

    def send_by_generator(self, url: str, generator: Generator, encoding: str = None) -> requests.Response:
        return self._send_by_generator(url, generator, encoding)

    def _send_by_generator(self, url: str, generator: Generator, encoding: str = None) -> requests.Response:
        content_type = "application/octet-stream"
        if encoding is not None:
            content_type = content_type + "; charset={}".format(encoding)

        headers = {"Content-Type": content_type}
        response = requests.post(url, data=generator, headers=headers)
        response.close()
        return response

    def __send(self, io_stream, chunk_size, on_chunk_send, total_length):
        from functools import partial
        bytes_send = 0
        for chunk in iter(partial(io_stream.read, chunk_size), b''):
            yield chunk
            bytes_send = bytes_send + len(chunk)
            if callable(on_chunk_send):
                on_chunk_send(bytes_send, total_length)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: August 27th 2020 13:10:16 pm
'''


from .constants import *
from ..util import BaseLocal, join_path

BaseLocal.import_csv([join_path(__path__[0], Localization_File)])



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 27th 2020

This module integrates Tornado Web Server as the base web framework.

Modified By: hsky77
Last Updated: August 27th 2020 15:32:05 pm
'''

from .mixin import TornadoMixin
from .base_controller import (RequestController, WebSocketController,
                              StreamingDownloadController, StreamingFileUploadController, StreamingUploadController)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

This module defines the base controller classes

Modified By: hsky77
Last Updated: August 27th 2020 12:31:57 pm
'''

from typing import Callable, Any, List

from tornado.web import RequestHandler

from ...util import DynamicTypeEnum
from .. import Controller_Module_Folder


class ControllerType(DynamicTypeEnum):
    """base abstract controller enum class"""
    @staticmethod
    def get_controller_enum_class() -> List[DynamicTypeEnum]:
        try:
            return DynamicTypeEnum.get_dynamic_class_enum_class(Controller_Module_Folder)
        except:
            pass

    def import_class(self):
        return super().import_class(cls_type=RequestHandler)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 27th 2020

Modified By: hsky77
Last Updated: August 27th 2020 16:00:06 pm
'''

import logging
from typing import Any, Callable, Union, Dict
from tornado.web import Application

from ..component import DefaultComponentTypes, create_server_component_manager
from ..component.default import LocalizationComponent, LoggerComponent, CallbackComponent, ExecutorComponent, ServicesComponent
from .. import LocalCode_Application_Closing, LocalCode_Application_Closed, Component_Module_Folder


class ComponentApplicationMixin():
    def init_application_mixin(self, settings: Dict[str, Any]) -> None:
        import threading
        self.do_close = False
        self.exiting = False

        if threading.current_thread() is threading.main_thread():
            import signal
            signal.signal(signal.SIGINT, self.signal_handler)

        if Component_Module_Folder in settings:
            self.component_manager = create_server_component_manager(
                settings[Component_Module_Folder],
                settings['root_dir'])
        else:
            # default component manager
            self.component_manager = create_server_component_manager(
                None, settings['root_dir'])

        self.component_manager.invoke(
            DefaultComponentTypes.Logger, 'update_default_logger', settings['debug'])

    def signal_handler(self, signum, frame):
        self.exit()

    def exit(self):
        self.logger.info(self.comp_local.get_message(
            LocalCode_Application_Closing))
        self.do_close = True

    async def do_exit(self):
        if self.do_close:
            if not self.exiting:
                self.exiting = True
                # await
                from tornado.ioloop import IOLoop
                IOLoop.current().stop()
                await self.component_manager.dispose_components()
                self.logger.info(self.comp_local.get_message(
                    LocalCode_Application_Closed))

    @property
    def comp_local(self) -> LocalizationComponent:
        """Get the LocalizationComponent singleton instance."""
        if not hasattr(self, '_comp_local'):
            self._comp_local = self.component_manager.get_component(
                DefaultComponentTypes.Localization)
        return self._comp_local

    @property
    def comp_logger(self) -> LoggerComponent:
        """Get the LoggerComponent singleton instance."""
        if not hasattr(self, '_comp_logger'):
            self._comp_logger = self.component_manager.get_component(
                DefaultComponentTypes.Logger)
        return self._comp_logger

    @property
    def comp_callbacks(self) -> CallbackComponent:
        """Get the CallbackComponent singleton instance."""
        if not hasattr(self, '_comp_callbacks'):
            self._comp_callbacks = self.component_manager.get_component(
                DefaultComponentTypes.Callback)
        return self._comp_callbacks

    @property
    def comp_executor(self) -> ExecutorComponent:
        """Get the ExecutorComponent singleton instance."""
        if not hasattr(self, '_comp_executor'):
            self._comp_executor = self.component_manager.get_component(
                DefaultComponentTypes.Executor)
        return self._comp_executor

    @property
    def comp_services(self) -> ServicesComponent:
        """Get the ServicesComponent singleton instance."""
        if not hasattr(self, '_comp_services'):
            self._comp_services = self.component_manager.get_component(
                DefaultComponentTypes.Service)
        return self._comp_services



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: September 5th 2020 23:18:04 pm
'''

import shutil
import setuptools
from hyssop import Version, __name__

shutil.rmtree('dist')
shutil.rmtree('build')

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=__name__,
    version=Version,
    author="hsky77",
    author_email="howardlkung@gmail.com",
    description="component-based architecture and project hierarchy based on opensource web framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hsky77/hyssop",
    packages=setuptools.find_packages(
        exclude=('hyssop_extension', 'hyssop_extension.*')),
    license="MIT License",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['tornado>=6.0.3',
                      'tornado-swagger>=1.2.4',
                      'PyYAML>=5.1.1',
                      'coloredlogs>=10.0',
                      'aiohttp>=3.6.2',
                      'requests>=2.20.0'],
    python_requires='>=3.6',
    package_data={'': ['*.yaml', '*.csv']}
)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: September 1st 2020 10:59:23 am
'''


import os
from typing import Awaitable, Dict


from tornado.web import url

from ..util import join_path, DynamicTypeEnum, join_to_abs_path, BaseLocal
from .component import (ComponentManager, ComponentTypes, DefaultComponentTypes,
                        create_server_component_manager)

from . import LocalCode_File_Not_Found, Hyssop_Web_Config_File


class HyssopServer():
    def start(self):
        from tornado.ioloop import IOLoop, PeriodicCallback
        PeriodicCallback(self.app.do_exit, 1000).start()
        IOLoop.current().start()

    def stop(self):
        if self.app:
            self.app.exit()

    def start_periodic_callback(self, awaitable_func: Awaitable, *args, interval: int = 1000, **kwargs) -> None:
        from tornado.ioloop import PeriodicCallback
        from inspect import iscoroutinefunction
        if iscoroutinefunction(awaitable_func):
            PeriodicCallback(lambda: self.__periodic_callback_decorator(
                awaitable_func, *args, **kwargs), interval).start()

    async def __periodic_callback_decorator(self, func, *args, **kwargs):
        if not self.app.exiting:
            await func(*args, **kwargs)

    def init(self, server_dir: str, use_http_server: bool = False, config: Dict = None, **kwargs) -> None:
        self.root_dir = join_to_abs_path(server_dir)
        self.config_path = join_to_abs_path(
            server_dir, Hyssop_Web_Config_File)

        if config:
            self.config = config
            self.config_path = None
        else:
            with open(self.config_path, 'r') as f:
                import yaml
                from .config_validator import WebConfigRootValidator
                validator = WebConfigRootValidator(
                    yaml.load(f, Loader=yaml.SafeLoader))
                self.config = validator.parameter

        self._make_app()

        if use_http_server or 'ssl' in self.config:
            from tornado.web import HTTPServer

            if 'ssl' in self.config:
                if not os.path.isfile(join_path(self.config['ssl']['crt'])):
                    raise FileNotFoundError(BaseLocal.get_message(
                        LocalCode_File_Not_Found, join_path(self.config['ssl']['crt'])))
                if not os.path.isfile(join_path(self.config['ssl']['key'])):
                    raise FileNotFoundError(BaseLocal.get_message(
                        LocalCode_File_Not_Found, join_path(self.config['ssl']['key'])))

                import ssl
                ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                ssl_ctx.load_cert_chain(join_path(self.config['ssl']['crt']),
                                        join_path(self.config['ssl']['key']))
                if 'ca' in self.config['ssl']:
                    if not os.path.isfile(join_path(self.config['ssl']['ca'])):
                        raise FileNotFoundError(BaseLocal.get_message(
                            LocalCode_File_Not_Found, join_path(self.config['ssl']['ca'])))
                    ssl_ctx.load_verify_locations(
                        join_path(self.config['ssl']['ca']))

                self.server = HTTPServer(self.app, ssl_options=ssl_ctx)
            else:
                self.server = HTTPServer(self.app)

            self.server.listen(self.config['port'])
        else:
            self.app.listen(self.config['port'])

    def _make_app(self):
        from .controller import get_controllers

        settings = {
            'root_dir': self.root_dir,
            'config_path': self.config_path,
            **self.config
        }

        controllers, settings = get_controllers(settings, self.root_dir)

        self.app = self._make_app_instance(controllers, settings)
        self.app.init_application_mixin(settings)

    def _setup_api_doc(self, controllers, settings):
        """
        setup swagger api doc urls, parameter "settings" is a dict that has "doc" like:

        {
            'doc': {
                'swagger_url': '/api/doc',
                'api_base_url': '/',
                'description': 'Swagger API definition',
                'api_version': '1.0.0',
                'title': 'Swagger API'
            }
        }
        """
        if 'doc' in settings:
            api_route = '/api/doc'
            description = 'Swagger API definition'
            api_version: str = '1.0.0'
            title = 'Swagger API'
            contact = ''

            if type(settings['doc']) is dict:
                api_route = settings['doc'].get('api_route', api_route)
                description = settings['doc'].get('description', description)
                api_version = settings['doc'].get('version', api_version)
                title = settings['doc'].get('title', title)
                contact = settings['doc'].get('contact', contact)

            from tornado_swagger.setup import setup_swagger
            controllers = [url(route, c, name=c.__name__)
                           for route, c, _ in controllers]
            setup_swagger(controllers, description=description,
                          api_version=api_version, title=title, contact=contact, swagger_url=api_route)

        return controllers

    def _make_app_instance(self, controllers, settings):
        if "framework" in settings:
            if settings["framework"] == "tornado":
                from .application.tornado import TornadoApplication
                from tornado_swagger.setup import setup_swagger
                controllers = self._setup_api_doc(controllers, settings)
                return TornadoApplication(controllers, **settings)
        else:
            from .application.tornado import TornadoApplication
            from tornado_swagger.setup import setup_swagger
            controllers = self._setup_api_doc(controllers, settings)
            return TornadoApplication(controllers, **settings)



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 21st 2020

Modified By: hsky77
Last Updated: September 24th 2020 12:13:59 pm
'''

import time
from unittest import TestCase
from typing import List, Dict, Any

from ..util import DynamicTypeEnum, Worker

from ..web.server import HyssopServer
from ..web.constants import Component_Module_Folder, Controller_Module_Folder


class UnitTestTypes(DynamicTypeEnum):
    """base abstract unitest enum class"""

    @staticmethod
    def get_unittest_enum_class() -> List[DynamicTypeEnum]:
        from ..web import Unittest_Module_Folder
        try:
            return DynamicTypeEnum.get_dynamic_class_enum_class(Unittest_Module_Folder)
        except:
            pass

    def import_class(self):
        return super().import_class(cls_type=UnitTestCase)


class UnitTestCase(TestCase):
    """hyssop unittest case abstract class"""

    def test(self):
        raise NotImplementedError()


class UnitTestServer():
    default_component_setting = {
        'localization': {'dir': None},
        'logger': {'dir': 'files/logs', 'log_to_resources': False, 'log_to_console': False},
        'executor': {'worker_count': 3},
        'services':
        {
            'async_connection_limit': 20,
            'async_connection_limit_pre_host': 8,
            'routes': {}
        }
    }

    def __init__(self, name: str = 'UT Server', port: int = 58564, debug: bool = True):
        self._config = {
            'name': name,
            'port': port,
            'debug': debug,
            Component_Module_Folder: self.default_component_setting,
            Controller_Module_Folder: {}
        }

        self.server_exception = None
        self._server = HyssopServer()

    @property
    def running(self):
        return self._running if hasattr(self, '_running') else False

    def validate_config(self, component_config: Dict[str, Any] = None, controller_config: [str, Any] = None) -> None:
        if component_config:
            from ..web.config_validator import WebConfigComponentValidator
            WebConfigComponentValidator(component_config)

        if controller_config:
            from ..web.config_validator import WebConfigControllerValidator
            WebConfigControllerValidator(controller_config)

    def set_config(self, dir_path: str, component_config: Dict[str, Any] = None, controller_config: [str, Any] = None) -> None:
        from ..web.config_validator import WebConfigValidator

        self._root_dir = dir_path

        if component_config:
            self._config[Component_Module_Folder] = {
                **self._config[Component_Module_Folder], **component_config}

        if controller_config:
            self._config[Controller_Module_Folder] = {
                **self._config[Controller_Module_Folder], **controller_config}

        WebConfigValidator(self._config)

    def start(self):
        self._worker = Worker('server_worker')
        self._worker.run_method(self._start, self._server, self._root_dir, self._config,
                                on_exception=self._on_worker_exception)

        while not self.running:
            if self.server_exception:
                raise self.server_exception from self.server_exception
            time.sleep(0.5)

    def stop(self):
        if self._server:
            self._server.stop()
            self._server = None
            self._running = False

        if self._worker:
            self._worker.dispose()
            while self._worker.is_func_running:  # wait for server stoped
                time.sleep(0)

            self._worker = None

    def _start(self, server: HyssopServer, root_dir: str, config: dict):
        from tornado.ioloop import IOLoop
        IOLoop().make_current()

        server.init(root_dir, config=config)
        self._running = True
        server.start()

    def _on_worker_exception(self, e: Exception):
        self._running = False
        self.server_exception = e



# Copyright (C) 2020-Present the hyssop authors and contributors.
#
# This module is part of hyssop and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

'''
File created: August 20th 2020

Modified By: hsky77
Last Updated: September 20th 2020 18:47:39 pm
'''

from typing import Dict, Any

from ...component import DefaultComponentTypes, ComponentManager
from ...component.default import CallbackComponent, LocalizationComponent, ServicesComponent, ExecutorComponent, LoggerComponent, BaseSyncLogger

from ...application.tornado import TornadoApplication


class TornadoMixin():
    """hyssop controller helper functions and variables"""

    @property
    def hyssop_application(self) -> TornadoApplication:
        return self.application

    @property
    def application_name(self) -> str:
        return self.hyssop_application.settings.get('name', 'Hyssop Server')

    @property
    def component_manager(self) -> ComponentManager:
        return self.hyssop_application.component_manager

    @property
    def root_dir(self) -> str:
        return self.hyssop_application.settings['root_dir']

    @property
    def debug(self) -> bool:
        return self.hyssop_application.settings['debug'] if 'debug' in self.hyssop_application.settings else False

    @property
    def type_name(self) -> str:
        return type(self).__name__

    @property
    def component_services(self) -> ServicesComponent:
        return self.hyssop_application.comp_services

    @property
    def component_callbacks(self) -> CallbackComponent:
        return self.hyssop_application.comp_callbacks

    @property
    def component_localization(self) -> LocalizationComponent:
        return self.hyssop_application.comp_local

    @property
    def component_executor(self) -> ExecutorComponent:
        return self.hyssop_application.comp_executor

    @property
    def logger(self) -> BaseSyncLogger:
        if not hasattr(self, '_logger'):
            self._logger = self.hyssop_application.comp_logger.get_logger(
                self.type_name)
        return self._logger

    def log_info(self, msg: str, *args, exc_info=None, extra=None, stack_info=False) -> None:
        self.hyssop_application.comp_executor.run_method_in_queue(
            self.logger.info, msg, *args, exc_info=exc_info, extra=extra, stack_info=stack_info)

    def log_warning(self, msg: str, *args, exc_info=None, extra=None, stack_info=False) -> None:
        self.hyssop_application.comp_executor.run_method_in_queue(
            self.logger.warning, msg, *args, exc_info=exc_info, extra=extra, stack_info=stack_info)

    def log_error(self, msg: str, *args, exc_info=None, extra=None, stack_info=False) -> None:
        self.hyssop_application.comp_executor.run_method_in_queue(
            self.logger.error, msg, *args, exc_info=exc_info, extra=extra, stack_info=stack_info)

    def get_arguments_dict(self, include_arguments=None) -> Dict[str, Any]:
        '''Return dict of arguments with key name in include_arguments or all of the arguments when include_arguments is None'''
        if include_arguments is not None:
            return {k: self.get_argument(k, default=None)
                    for k in self.request.arguments
                    if k in include_arguments}
        else:
            return {k: self.get_argument(k, default=None)
                    for k in self.request.arguments}
